@@ -17,8 +17,8 @@ main = do
 
 updateCanvas :: DrawingArea -> PangoLayout -> Event -> IO Bool
 updateCanvas canvas text (Expose { eventArea=rect }) = do
-  win <- drawingAreaGetDrawWindow canvas
-  (width,height) <- drawingAreaGetSize canvas
+  win <- widgetGetDrawWindow canvas
+  (width,height) <- widgetGetSize canvas
   gc <- gcNew win
   gcSetValues gc $ newGCValues {
     foreground = Color 65535 0 0,
