@@ -32,6 +32,10 @@ module Graphics.UI.Gtk.ModelView.TreeStore (
 -- * Constructors
   treeStoreNew,
 
+-- * Implementation of Interfaces
+  treeStoreDefaultDragSourceIface,
+  treeStoreDefaultDragDestIface,
+
 -- * Methods
   treeStoreGetValue,
   treeStoreGetTree,
@@ -54,10 +58,11 @@ import Control.Monad ( liftM, when )
 import Control.Exception (assert)
 import Data.IORef
 import Graphics.UI.Gtk.ModelView.Types
-import Graphics.UI.Gtk.Types (GObjectClass, TreeModelClass)
-import Graphics.UI.Gtk.TreeList.TreePath (TreePath)
+import Graphics.UI.Gtk.Types (GObjectClass(..), TreeModelClass)
 import Graphics.UI.Gtk.ModelView.CustomStore
-import Graphics.UI.Gtk.TreeList.TreeIter
+import Graphics.UI.Gtk.ModelView.TreeModel
+import Graphics.UI.Gtk.ModelView.TreeDrag
+import Control.Monad.Trans ( liftIO )
 
 --------------------------------------------
 -- internal model data types
@@ -67,9 +72,11 @@ import Graphics.UI.Gtk.TreeList.TreeIter
 --
 newtype TreeStore a = TreeStore (CustomTreeModel (IORef (Store a)) a)
 
-instance GObjectClass (TreeStore a)
-instance TreeModelClass (TreeStore a)
 instance TypedTreeModelClass TreeStore
+instance TreeModelClass (TreeStore a)
+instance GObjectClass (TreeStore a) where
+  toGObject (TreeStore tm) = toGObject tm
+  unsafeCastGObject = TreeStore . unsafeCastGObject
 
 -- | Maximum number of nodes on each level.
 --
@@ -90,7 +97,20 @@ data Store a = Store {
 --   list. Each 'Tree' in the forest corresponds to one top-level node.
 --
 treeStoreNew :: Forest a -> IO (TreeStore a)
-treeStoreNew forest = do
+treeStoreNew forest = treeStoreNewDND forest
+                        (Just treeStoreDefaultDragSourceIface)
+                        (Just treeStoreDefaultDragDestIface)
+
+-- | Create a new list store.
+--
+-- * In addition to 'treeStoreNew', this function takes an two interfaces
+--   to implement user-defined drag-and-drop functionality.
+--
+treeStoreNewDND :: Forest a -- ^ the inital tree stored in this model
+  -> Maybe (DragSourceIface TreeStore a) -- ^ an optional interface for drags
+  -> Maybe (DragDestIface TreeStore a) -- ^ an optional interface to handle drops
+  -> IO (TreeStore a)
+treeStoreNewDND forest mDSource mDDest = do
   storeRef <- newIORef Store {
       depth = calcForestDepth forest,
       content = storeToCache forest
@@ -102,51 +122,90 @@ treeStoreNew forest = do
         writeIORef storeRef store { content = cache' }
         return result
 
-  liftM TreeStore $ customTreeModelNew storeRef CustomTreeModelImplementation {
-    customTreeModelGetFlags = return [],
+  customTreeModelNew storeRef TreeStore TreeModelIface {
+    treeModelIfaceGetFlags = return [],
 
-    customTreeModelGetIter = \path -> withStore $
+    treeModelIfaceGetIter = \path -> withStore $
       \Store { depth = d } -> fromPath d path,
 
-    customTreeModelGetPath = \iter -> withStore $
+    treeModelIfaceGetPath = \iter -> withStore $
       \Store { depth = d } -> toPath d iter,
 
-    customTreeModelGetRow  = \iter -> withStoreUpdateCache $
+    treeModelIfaceGetRow  = \iter -> withStoreUpdateCache $
       \Store { depth = d, content = cache } ->
         case checkSuccess d iter cache of
           (True, cache'@((_, (Node { rootLabel = val }:_)):_)) ->
             (val, cache')
           _ -> error "TreeStore.getRow: iter does not refer to a valid entry",
 
-    customTreeModelIterNext = \iter -> withStoreUpdateCache $
+    treeModelIfaceIterNext = \iter -> withStoreUpdateCache $
       \Store { depth = d, content = cache } -> iterNext d iter cache,
 
-    customTreeModelIterChildren = \mIter -> withStoreUpdateCache $
+    treeModelIfaceIterChildren = \mIter -> withStoreUpdateCache $
       \Store { depth = d, content = cache } ->
       let iter = fromMaybe invalidIter mIter
        in iterNthChild d 0 iter cache,
 
-    customTreeModelIterHasChild = \iter -> withStoreUpdateCache $
+    treeModelIfaceIterHasChild = \iter -> withStoreUpdateCache $
       \Store { depth = d, content = cache } ->
        let (mIter, cache') = iterNthChild d 0 iter cache
         in (isJust mIter, cache'),
 
-    customTreeModelIterNChildren = \mIter -> withStoreUpdateCache $
+    treeModelIfaceIterNChildren = \mIter -> withStoreUpdateCache $
       \Store { depth = d, content = cache } ->
       let iter = fromMaybe invalidIter mIter
        in iterNChildren d iter cache,
 
-    customTreeModelIterNthChild = \mIter idx  -> withStoreUpdateCache $
+    treeModelIfaceIterNthChild = \mIter idx  -> withStoreUpdateCache $
       \Store { depth = d, content = cache } ->
       let iter = fromMaybe invalidIter mIter
        in iterNthChild d idx iter cache,
 
-    customTreeModelIterParent = \iter -> withStore $
+    treeModelIfaceIterParent = \iter -> withStore $
       \Store { depth = d } -> iterParent d iter,
 
-    customTreeModelRefNode = \_ -> return (),
-    customTreeModelUnrefNode = \_ -> return ()
-   }
+    treeModelIfaceRefNode = \_ -> return (),
+    treeModelIfaceUnrefNode = \_ -> return ()
+   } mDSource mDDest
+
+
+-- | Default drag functions for
+-- 'Graphics.UI.Gtk.ModelView.TreeStore'. These functions allow the rows of
+-- the model to serve as drag source. Any row is allowed to be dragged and the
+-- data set in the 'SelectionDataM' object is set with 'treeSetRowDragData',
+-- i.e. it contains the model and the 'TreePath' to the row.
+treeStoreDefaultDragSourceIface :: DragSourceIface TreeStore row
+treeStoreDefaultDragSourceIface = DragSourceIface {
+    treeDragSourceRowDraggable = \_ _-> return True,
+    treeDragSourceDragDataGet = treeSetRowDragData,
+    treeDragSourceDragDataDelete = \model dest@(_:_) -> do
+            liftIO $ treeStoreRemove model dest
+            return True
+
+  }
+
+-- | Default drop functions for 'Graphics.UI.Gtk.ModelView.TreeStore'. These
+--   functions accept a row and insert the row into the new location if it is
+--   dragged into a tree view
+-- that uses the same model.
+treeStoreDefaultDragDestIface :: DragDestIface TreeStore row
+treeStoreDefaultDragDestIface = DragDestIface {
+    treeDragDestRowDropPossible = \model dest -> do
+      mModelPath <- treeGetRowDragData
+      case mModelPath of
+        Nothing -> return False
+        Just (model', source) -> return (treeModelEqual model model'),
+    treeDragDestDragDataReceived = \model dest@(_:_) -> do
+      mModelPath <- treeGetRowDragData
+      case mModelPath of
+        Nothing -> return False
+        Just (model', source@(_:_)) ->
+          if not (treeModelEqual model model') then return False
+          else liftIO $ do
+            row <- treeStoreGetTree model source
+            treeStoreInsertTree model (init dest) (last dest) row
+            return True
+  }
 
 --------------------------------------------
 -- low level bit-twiddling utility functions
@@ -413,13 +472,15 @@ treeStoreInsertForest (TreeStore model) path pos nodes = do
            (idx, toggle))
   Store { depth = depth } <- readIORef (customTreeModelGetPrivate model)
   let rpath = reverse path
+  stamp <- customTreeModelGetStamp model
   sequence_ [ let p' = reverse p
                   Just iter = fromPath depth p'
-               in treeModelRowInserted model p' iter
+               in treeModelRowInserted model p' (treeIterSetStamp iter stamp)
             | (i, node) <- zip [idx..] nodes
             , p <- paths (i : rpath) node ]
   let Just iter = fromPath depth path
-  when toggle $ treeModelRowHasChildToggled model path iter
+  when toggle $ treeModelRowHasChildToggled model path
+                (treeIterSetStamp iter stamp)
 
   where paths :: TreePath -> Tree a -> [TreePath]
         paths path Node { subForest = ts } =
@@ -565,7 +626,8 @@ treeStoreChangeM (TreeStore model) path act = do
 				 content = storeToCache newForest }, True)
   writeIORef (customTreeModelGetPrivate model) store'
   let Just iter = fromPath d path
-  when found $ treeModelRowChanged model path iter
+  stamp <- customTreeModelGetStamp model
+  when found $ treeModelRowChanged model path (treeIterSetStamp iter stamp)
   return found
 
 -- | Change a node in the forest.
